#include "camera_decoder.h"
#include "hi_dvpp_vdec.h"
#include "hi_media_common.h"
#include "hi_mpi_sys.h"

#include <iostream>

CameraDecoder::CameraDecoder(int channel_number) : m_channel_number(channel_number){}

CameraDecoder::~CameraDecoder()
{
    if (!m_is_initialized)
        return;
    hi_mpi_vdec_stop_recv_stream(m_channel_number);
    hi_mpi_vdec_destroy_chn(m_channel_number);
    hi_mpi_sys_exit();
}

bool CameraDecoder::initializeHiMpiSystem()
{
    int32_t ret = hi_mpi_sys_init();
    if (ret) {
        log_.LogError() << "Hi Mpi init failed with code: " << ret;
        return false;
    }

    //Create channel
    hi_vdec_chn_attr chnAttr;
    hi_pic_buf_attr  buf_attr;
    buf_attr.width         = m_width;
    buf_attr.height        = m_height;
    buf_attr.align         = 0;
    buf_attr.bit_width     = HI_DATA_BIT_WIDTH_8;
    buf_attr.pixel_format  = HI_PIXEL_FORMAT_YUV_SEMIPLANAR_420;
    buf_attr.compress_mode = HI_COMPRESS_MODE_NONE;

    chnAttr.type                       = HI_PT_H265;
    chnAttr.mode                       = HI_VDEC_SEND_MODE_FRAME;
    chnAttr.pic_width                  = m_width;
    chnAttr.pic_height                 = m_height;
    chnAttr.stream_buf_size            = m_output_size;
    chnAttr.frame_buf_cnt              = 16;
    chnAttr.frame_buf_size             = hi_vdec_get_pic_buf_size(HI_PT_H265, &buf_attr);
    // chnAttr.video_attr.ref_frame_num   = 12;
    chnAttr.video_attr.ref_frame_num   = 16;
    chnAttr.video_attr.temporal_mvp_en = HI_TRUE;
    chnAttr.video_attr.tmv_buf_size    = hi_vdec_get_tmv_buf_size(HI_PT_H265, m_width, m_height);

    ret = hi_mpi_vdec_create_chn(m_channel_number, &chnAttr);
    if (ret) {
    	log_.LogError() << "create channel_" << m_channel_number << " failed with code: " << ret;
        return false;
    }
    hi_vdec_chn_param chnParam;
    ret = hi_mpi_vdec_get_chn_param(m_channel_number, &chnParam);
    if (ret) {
    	log_.LogError() << "get chn param failed with code: " << ret;
        return false;
    }
    chnParam.video_param.dec_mode      = HI_VIDEO_DEC_MODE_IPB;
    chnParam.video_param.compress_mode = HI_COMPRESS_MODE_NONE;
    chnParam.video_param.video_format  = HI_VIDEO_FORMAT_LINEAR;
    chnParam.display_frame_num         = 3;
    chnParam.video_param.out_order     = HI_VIDEO_OUT_ORDER_DEC;
    chnParam.type                      = HI_PT_H265;

    ret = hi_mpi_vdec_set_chn_param(m_channel_number, &chnParam);
    if (ret) {
    	log_.LogError() << "set chn param failed with code: " << ret;
    	return false;
    }
    log_.LogInfo() << "Hi Mpi system initialized successfully";
    return true;
}

CameraDecoder::ErrorCode CameraDecoder::decode(uint8_t* input_frame_buffer, size_t input_frame_size, bool is_IDR_frame, uint8_t *decoded_frame_data, size_t *decoded_data_size)
{
    ErrorCode error_code = Error_Decode_Failed;
    if (!m_is_initialized) {
        if (!initializeHiMpiSystem())
            return Error_Initialization;
        m_is_initialized = true;
    }
    if (!m_first_IDR_received) { //First of GoP(Group of pictures)
        log_.LogInfo() << "waiting for IDR....." << is_IDR_frame;
        if (is_IDR_frame)
        {
            m_first_IDR_received = true;
        }
        else
            return Error_IDR_Not_Received;
    } //we got the first IDR frame and we can do the decoding

    auto ret = hi_mpi_dvpp_malloc(0, (hi_void**) &m_inputAddr, input_frame_size);
    if (ret) {
        log_.LogError() << "Get input buffer for decoder failed  with code: " << ret;
        return Error_Allocation_Failed;
    }

    memcpy(m_inputAddr, input_frame_buffer, input_frame_size);

    ret = hi_mpi_dvpp_malloc(0, (hi_void**) &m_outputAddr, m_output_size);
    if (ret) {
        log_.LogError() << "Get output buffer for decoder failed with code: " << ret;
        ret = hi_mpi_dvpp_free(m_inputAddr);
        if (ret)
            log_.LogError() << "Free input buffer failed with code: " << ret;
        return Error_Allocation_Failed;
    }

    // Start to receive the code stream
    ret = hi_mpi_vdec_start_recv_stream(m_channel_number);
    if (ret) {
        log_.LogError() << "start recv stream failed with code: " << ret;
        ret = hi_mpi_dvpp_free(m_inputAddr);
        if (ret)
            log_.LogError() << "Free input buffer failed with code: " << ret;
        ret = hi_mpi_dvpp_free(m_outputAddr);
        if (ret)
            log_.LogError() << "Free output buffer failed with code: " << ret;
        return Error_Start_Receive_Failed;
    }

    hi_vdec_stream stream;
    stream.addr          = m_inputAddr;
    stream.len           = input_frame_size;
    stream.end_of_frame  = HI_TRUE;
    stream.end_of_stream = HI_FALSE;
    stream.need_display  = HI_TRUE;

    //Construct a structure to store one frame of output result information
    hi_vdec_pic_info outPicInfo;
    outPicInfo.width         = m_width;
    outPicInfo.height        = m_height;
    outPicInfo.width_stride  = m_width;
    outPicInfo.height_stride = m_height;
    outPicInfo.pixel_format  = HI_PIXEL_FORMAT_YUV_SEMIPLANAR_420;
    outPicInfo.vir_addr      = (uint64_t) m_outputAddr;
    outPicInfo.buffer_size   = m_output_size;

    //Send a frame of code stream
    ret = hi_mpi_vdec_send_stream(m_channel_number, &stream, &outPicInfo, 0);
    if (ret) {
        log_.LogError() << "Send stream failed with code: " << ret;
        ret = hi_mpi_dvpp_free(m_inputAddr);
        if (ret)
            log_.LogError() << "Free input buffer failed with code: " << ret;
        ret = hi_mpi_dvpp_free(m_outputAddr);
        if (ret)
            log_.LogError() << "Free output buffer failed with code: " << ret;
        return Error_Send_Stream_Failed;
    }

    hi_video_frame_info     frame;
    hi_vdec_supplement_info stSupplement;
    ret = hi_mpi_vdec_get_frame(m_channel_number, &frame, &stSupplement, &stream, 40);
    if (ret) {
        log_.LogError() << "Get frame failed with code: " << ret;
        return Error_Get_Frame_Failed;
    }
    auto decResult = frame.v_frame.frame_flag;
    if (decResult == 0) { // 0: Decode success
        memcpy(decoded_frame_data, frame.v_frame.virt_addr[0], m_output_size);
        // decoded_frame_data = (uint8_t *)frame.v_frame.virt_addr[0];
        *decoded_data_size = m_output_size;
        error_code = Error_None;
    } else if (decResult == 1) { // 1: Decode fail
        log_.LogError() << "Chn " << m_channel_number << " GetFrame Success, Decode Fail";
        error_code = Error_Decode_Failed;
    } else if (decResult == 2) { // 2:This result is returned for the second field of
        log_.LogError() << "Chn " << m_channel_number << " GetFrame Success, No Picture";
        error_code = Error_Decode_Failed;
    } else if (decResult == 3) { // 3: Reference frame number set error_code
        log_.LogError() << "Chn " << m_channel_number << " GetFrame Success, RefFrame Num Error";
        error_code = Error_Decode_Failed;
    } else if (decResult == 4) { // 4: Reference frame size set error_code
        log_.LogError() << "Chn " << m_channel_number << " GetFrame Success, RefFrame Size Error";
        error_code = Error_Decode_Failed;
    }
    ret = hi_mpi_dvpp_free(frame.v_frame.virt_addr[0]);
    if (ret)
        log_.LogError() << "Free frame failed with code: " << ret;
//    ret = hi_mpi_dvpp_free(stream.addr);
//    if (ret)
//		log_.LogError() << "Free stream failed with code: " << ret;
    ret = hi_mpi_vdec_release_frame(m_channel_number, &frame);
    if (ret)
        log_.LogError() << "Release frame failed with code: " << ret;
    return error_code;
} // CameraDecoder::decode
